/*
This file is part of web2w.
Copyright 2017 Martin Ruckert

web2w is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

web2w is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with web2w.  If not, see <http://www.gnu.org/licenses/>.
 
Martin Ruckert, Hochschule Muenchen, Lothstrasse 64, 80336 Muenchen
*/
% This file is part of HINT
% Copyright 2017 Martin Ruckert
% 
% HINT is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% HINT is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with HINT.  If not, see <http://www.gnu.org/licenses/>.
%  
% Martin Ruckert, Hochschule Muenchen, Lothstrasse 64, 80336 Muenchen

% This is ktex.ch, a cweb change file containing code to extend TeX
% to conform to the conventions of the implementations of TeX distributed
% as TeX Live.
%
% This change file is copyright (C) 2021 by Martin Ruckert;
% all rights are reserved. Copying of this file is authorized only if
% or if you make absolutely no changes to your copy. 
% Programs such as TIE allow the application of several change
% files to tex.web; the master files tex.web and texlive.ch should stay intact.
%
% The TeX program is copyright (C) 1982 by D. E. Knuth.
% TeX is a trademark of the American Mathematical Society.
%
% This applying this change file to tex.w should produce a program
% that passes the triptest if tex.w does so.
%
% Note that this change file is a cweb change file. You can not apply it
% to tex.web but you must apply it to tex.w which can be obtained using
% the web2w translator.

\section{In Limbo}
Before we enter the first section, we redefine the title.
@x
\def\title{\eTeX}
@y
\def\title{k\TeX}
@z

\section{ini\TeX}
\TeX's table entries should not be initialized unless |iniversion| is true.
Especially because |format_ident| would be set to |"(INITEX)"| 
which would later prevent the loading of a format file. 
%line 420
@x 
the codewords `$|init|\ldots|tini|$'.

@<Initialize whatever...@>=
@<Set initial values of key variables@>@/
#ifdef @!INIT
@<Initialize table entries (done by \.{INITEX} only)@>@;
@y becomes
|#ifdef| |INIT|\dots\ |#endif|.
 
\TeX\ Live has established the common practice
to select the initialization code at runtime
using the |iniversion| variable.

@<Initialize whatever...@>=
@<Set initial values of key variables@>@/
#ifdef @!INIT
if (iniversion) {@+@<Initialize table entries (done by \.{INITEX} only)@>@;@+}
@z 

\section{Character set}
The \TeX\ processor converts between ASCII code and
the user's external character set by means of arrays |xord| and |xchr|
that are analogous to \Pascal/'s |ord| and |chr| functions.
We change the definitions of |xchr| to the identity mapping like in \TeX\ Live.
Therefore we replace 
%line 728 
@x 
for (i=0; i<=037; i++) xchr[i]=' ';
for (i=0177; i<=0377; i++) xchr[i]=' ';
@y  by 
for (i=0; i<=037; i++) xchr[i]=chr(i); /* k\TeX\ */
for (i=0177; i<=0377; i++) xchr[i]=chr(i); /* k\TeX\ */
@z 

\section{Opening input files}
To open files, \TeX\ uses |reset|.
Now we use the {\tt kpathsearch} library to implement new functions.
To do so we reduce \TeX\ functions to prototypes and postpone the 
actual implementation.
%line 887
@x
@ The \ph\ compiler with which the present version of \TeX\ was prepared has
extended the rules of \PASCAL\ in a very convenient way. To open file~|f|,
we can write
$$\vbox{\halign{#\hfil\qquad&#\hfil\cr
|reset(f,@t\\{name}@>,"/O")|&for input;\cr
|rewrite(f,@t\\{name}@>,"/O")|&for output.\cr}}$$
The `\\{name}' parameter, which is of type `{\bf packed array
$[\langle\\{any}\rangle]$ of \\{char}}', stands for the name of
the external file that is being opened for input or output.
Blank spaces that might appear in \\{name} are ignored.

The `\.{/O}' parameter tells the operating system not to issue its own
error messages if something goes wrong. If a file of the specified name
cannot be found, or if such a file cannot be opened for some other reason
(e.g., someone may already be trying to write the same file), we will have
|@!erstat(f)!=0| after an unsuccessful |reset| or |rewrite|.  This allows
\TeX\ to undertake appropriate corrective action.
@:PASCAL H}{\ph@>
@^system dependencies@>

\TeX's file-opening procedures return |false| if no file identified by
|name_of_file| could be opened.

@d reset_OK(A) erstat(A)==0
@d rewrite_OK(A) erstat(A)==0

@p static bool a_open_in(alpha_file *f)
   /*open a text file for input*/
{@+reset((*f), name_of_file,"r");return reset_OK((*f));
}
@#
static bool a_open_out(alpha_file *f)
   /*open a text file for output*/
{@+rewrite((*f), name_of_file,"w");return rewrite_OK((*f));
}
@#
static bool b_open_in(byte_file *f)
   /*open a binary file for input*/
{@+reset((*f), name_of_file,"rb");return reset_OK((*f));
}
@#
static bool b_open_out(byte_file *f)
   /*open a binary file for output*/
{@+rewrite((*f), name_of_file,"wb");return rewrite_OK((*f));
}
@#
static bool w_open_in(word_file *f)
   /*open a word file for input*/
{@+reset((*f), name_of_file,"rb");return reset_OK((*f));
}
@#
#ifdef @!INIT
static bool w_open_out(word_file *f)
   /*open a word file for output*/
{@+rewrite((*f), name_of_file,"wb");return rewrite_OK((*f));
}
#endif
@y becomes
@ k\TeX\ uses the {\tt kpathsearch} library to implement access to files.
To do so we declare \TeX's functions here and postpone the 
actual implementations. These functions do not to issue their own
error messages if something goes wrong. If a file identified by
|name_of_file| cannot be found, 
or if such a file cannot be opened for some other reason
(e.g., someone may already be trying to write the same file)
\TeX's file-opening procedures return |false|.
This allows \TeX\ to undertake appropriate corrective action.
@^system dependencies@>

@p static bool a_open_in(alpha_file *f);   /*open a text file for input*/
static bool a_open_out(alpha_file *f); /*open a text file for output*/
static bool b_open_in(byte_file *f); /*open a binary file for input*/
static bool b_open_out(byte_file *f); /*open a binary file for output*/
static bool w_open_in(word_file *f); /*open a word file for input*/
#ifdef @!INIT
static bool w_open_out(word_file *f); /*open a word file for output*/
#endif
@z

\section{The command line}
We call the function 
|input_command_line| immediately after calling |t_open_in|.
%line 1069
@x 
@ The following program does the required initialization
without retrieving a possible command line.
It should be clear how to modify this routine to deal with command lines,
if the system permits them.
@^system dependencies@>

@p static bool init_terminal(void) /*gets the terminal input started*/
{@+
t_open_in;
@y  becomes 
@ The following program calls |input_command_line| 
to retrieve a possible command line.
@^system dependencies@>

@p static bool init_terminal(void) /*gets the terminal input started*/
{@+
t_open_in;
if (input_command_line()) return true; /* k\TeX\ */
@z 

\section{Output files}
We do no longer need  \TeX's |reset| and |rewrite| macros.
%line 1565
@x 
#define reset(F,N,M)   @[((F).f=fopen((char *)(N)+1,M),\
                                 (F).f!=NULL?get(F):0)@]
#define rewrite(F,N,M) @[((F).f=fopen((char *)(N)+1,M))@]
@y
@z 

If there is no |format_ident|, we use |dump_name| as the format name.

%line 1702
@x 
@<Initialize the output...@>=
wterm("%s",banner);
if (format_ident==0) wterm_ln(" (no format preloaded)");
@y  becomes 
k\TeX, according to the conventions of \TeX\ Live,
 prints the |dump_name| if no format identifier is known.
@<Initialize the output...@>=
wterm("%s",banner);
if (format_ident==0) wterm_ln(" (preloaded format=%s)", dump_name); 
@z 

Next, we modify the |print_err| function.
We call the function |print_file_line|
if the option |filelineerrorstylep| is set
instead of calling |print_nl("! ")|. 

%line 1793
@x BEGIN
{@+if (interaction==error_stop_mode) wake_up_terminal;
  print_nl("! ");print(s);
@y END becomes BEGIN
{@+if (interaction==error_stop_mode) wake_up_terminal;
  if (filelineerrorstylep) print_file_line(); /* k\TeX\ */
  else print_nl("! ");
  print(s);
@z END

\section{Setting time and date}
We use the |tl_now| function to implement |fix_data_and_time|,
%line 5295
@x
Since standard \PASCAL\ cannot provide such information, something special
is needed. The program here simply specifies July 4, 1776, at noon; but
users probably want a better approximation to the truth.

@p static void fix_date_and_time(void)
{@+time=12*60; /*minutes since midnight*/ 
day=4; /*fourth day of the month*/ 
month=7; /*seventh month of the year*/ 
year=1776; /*Anno Domini*/ 
} 
@y  now becomes 
k\TeX\ calls |tl_now| to obtain the current time as a |tm| structure.
@p static void fix_date_and_time(void)
{@+struct tm *gmt=tl_now();
  time= gmt->tm_hour*60+gmt->tm_min;
  day= gmt->tm_mday;
  month=gmt->tm_mon+1;
  year=gmt->tm_year+1900;
}
@z 

\section{Source file names}


We initialize the |source_filename_stack| in the |begin_file_reading| function.
@x 
incr(in_open);push_input;index=in_open;
@y  becomes 
incr(in_open);push_input;index=in_open;@/
source_filename_stack[index]=0; /* k\TeX\ */
@z 



\section{Opening the format file}

We remove from |pack_buffered_name| the code
that adds the default area, possibly the default name,
and the default extension for format files to the
name contained in |buffer|. The \.{kpathsearch}
library will take care of these parts.

%line 10559
@x 
@ A messier routine is also needed, since format file names must be scanned
before \TeX's string mechanism has been initialized. We shall use the
global variable |TEX_format_default| to supply the text for default system areas
and extensions related to format files.
@^system dependencies@>

@d format_default_length 20 /*length of the |TEX_format_default| string*/
@d format_area_length 11 /*length of its area part*/
@d format_ext_length 4 /*length of its `\.{.fmt}' part*/
@d format_extension ".fmt" /*the extension, as a \.{WEB} constant*/

@<Glob...@>=
static unsigned char @!TEX_format_default0[format_default_length]="TeXformats/plain.fmt", *const @!TEX_format_default = @!TEX_format_default0-1;
@.TeXformats@>
@.plain@>
@^system dependencies@>

@ The static initialization of |TEX_format_default| 
makes a dynamic initialization obsolete.

@ @<Check the ``constant'' values for consistency@>=
if (format_default_length > file_name_size) bad=31;

@ Here is the messy routine that was just mentioned. It sets |name_of_file|
from the first |n| characters of |TEX_format_default|, followed by
|buffer[a dotdot b]|, followed by the last |format_ext_length| characters of
|TEX_format_default|.

We dare not give error messages here, since \TeX\ calls this routine before
the |error| routine is ready to roll. Instead, we simply drop excess characters,
since the error will be detected in another way when a strange file name
isn't found.
@^system dependencies@>

@p static void pack_buffered_name(small_number @!n, int @!a, int @!b)
{@+int k; /*number of positions filled in |name_of_file|*/ 
ASCII_code @!c; /*character being packed*/ 
int @!j; /*index into |buffer| or |TEX_format_default|*/ 
if (n+b-a+1+format_ext_length > file_name_size) 
  b=a+file_name_size-n-1-format_ext_length;
k=0;
for (j=1; j<=n; j++) append_to_name(xord[TEX_format_default[j]]);
for (j=a; j<=b; j++) append_to_name(buffer[j]);
for (j=format_default_length-format_ext_length+1; j<=format_default_length; j++) 
  append_to_name(xord[TEX_format_default[j]]);
if (k <= file_name_size) name_length=k;@+else name_length=file_name_size;
name_of_file[name_length+1]=0;
} 
@y 

@ The global variable |TEX_format_default| is no longer needed 
to supply the text for default system areas
and extensions related to format files.

@ Consequently there is no initialization of |TEX_format_default| either.

@ And there is no need to check the length of |TEX_format_default|.

@ The |format_extension|, however, is needed later on
to create the format name from the job name.

@d format_extension ".fmt" 
@z 

Similarly, we can remove these complications from the function |open_fmt_file|.
%line 10589
@x
@ Here is the only place we use |pack_buffered_name|. This part of the program
becomes active when a ``virgin'' \TeX\ is trying to get going, just after
the preliminary initialization, or when the user is substituting another
format file by typing `\.\&' after the initial `\.{**}' prompt.  The buffer
contains the first line of input in |buffer[loc dotdot(last-1)]|, where
|loc < last| and |buffer[loc]!=' '|.

@<Declare the function called |open_fmt_file|@>=
static bool open_fmt_file(void)
{@+
int j; /*the first space after the format file name*/
j=loc;
if (buffer[loc]=='&')
  {@+incr(loc);j=loc;buffer[last]=' ';
  while (buffer[j]!=' ') incr(j);
  pack_buffered_name(0, loc, j-1); /*try first without the system file area*/
  if (w_open_in(&fmt_file)) goto found;
  pack_buffered_name(format_area_length, loc, j-1);
     /*now try the system format file area*/
  if (w_open_in(&fmt_file)) goto found;
  wake_up_terminal;
  wterm_ln("Sorry, I can't find that format; will try PLAIN.");
@.Sorry, I can't find...@>
  update_terminal;
  }
   /*now pull out all the stops: try for the system \.{plain} file*/
pack_buffered_name(format_default_length-format_ext_length, 1, 0);
if (!w_open_in(&fmt_file))
  {@+wake_up_terminal;
  wterm_ln("I can't find the PLAIN format file!");
@.I can't find PLAIN...@>
@.plain@>
  return false;
  }
found: loc=j;return true;
}
@y
@ This part of the program
becomes active when a ``virgin'' \TeX\ is trying to get going, just after
the preliminary initialization, or when the user is substituting another
format file by typing `\.\&' after the initial `\.{**}' prompt.  The buffer
contains the first line of input in |buffer[loc dotdot(last-1)]|, where
|loc < last| and |buffer[loc]!=' '|.

k\TeX\ uses the {\tt kpathsearch} library to implement access to files.
To do so we declare |open_fmt_file| here and postpone the 
actual implementation.

@<Declare the function called |open_fmt_file|@>=
static bool open_fmt_file(void);
@z

Since k\TeX\ allows the specification of a job name
on the command line, the assignment to |job_name| needs to check
for it before using the default.
%line 10759
@x 
if (job_name==0) job_name=s_no("texput");
@y  becomes 
if (job_name==0) job_name=s_no(c_job_name?c_job_name:"texput"); /* k\TeX\ */
@z 

Adding a default file location can be removed:
%line 10817
@x 
  if (cur_area==0||cur_area[0]==0) 
    {@+pack_file_name(cur_name, TEX_area, cur_ext);
    if (a_open_in(&cur_file)) goto done;
    } 
@y 
@z

In the function |start_input|, we need to update the |source_filename_stack|.
%line 10824
@x 
done: name=a_make_name_string(&cur_file);
@y  becomes 
done: name=a_make_name_string(&cur_file);@/
source_filename_stack[in_open]=name; /* k\TeX\ */
@z 


The function |start_input| contains an other assignment to |job_name|
which needs to be replaced:
%line 10827
@x 
  {@+job_name=cur_name;open_log_file();
@y  becomes 
  {@+if (c_job_name==NULL) job_name=cur_name; 
     else job_name=s_no(c_job_name); open_log_file(); /* k\TeX\ */
@z 


When \TeX\ opens a font metric file, it used to supply the
|TEX_font_area|. The \.{kpathsearch} library does a better job.
%line 11015
@x 
if (aire[0]==0) pack_file_name(nom, TEX_font_area,".tfm");
else pack_file_name(nom, aire,".tfm");
@y  is simplified to:
pack_file_name(nom, "",".tfm"); /* k\TeX\ */
@z 

\section{Initialization}
Last not least, we have to insert the k\TeX\ initialization
routine.
%line 24292
@x
@p int main(void) {@! /*|start_here|*/ 
@y  becomes
@p
int main(int argc, char *argv[]) {@! /*|start_here|*/ 
  main_init(argc, argv); /* k\TeX\ */
@z 

%line 25045
@x
#ifdef @!INIT
init_prim(); /*call |primitive| for each primitive*/
init_str_ptr=str_ptr;init_pool_ptr=pool_ptr;fix_date_and_time();
#endif
@y
#ifdef @!INIT
if (iniversion)
{ init_prim(); /*call |primitive| for each primitive*/
  init_str_ptr=str_ptr;init_pool_ptr=pool_ptr;fix_date_and_time(); }
#endif
@z



There are even some changes to the code of $\epsilon$-\TeX. The
command line option \.{-etex} can substitute for the leading |'*'|
in the input.
@x
if ((buffer[loc]=='*')&&str_eq_str(format_ident," (INITEX)"))
@y becomes
if (iniversion && (buffer[loc]=='*'||etexp)&&str_eq_str(format_ident," (INITEX)"))
@z
and

@x
  incr(loc);eTeX_mode=1; /*enter extended mode*/
@y
  if (buffer[loc]=='*') incr(loc);
  eTeX_mode=1; /*enter extended mode*/
@z

\section{The \TeX\ Live extensions}
The additional code that belongs to \TeX\ live is inserted just before the index.

@x
@* Index.
@y
@* \TeX\ Live Integration.
A \TeX\ engine that aspires 
to become a member of the \TeX\ Live family of programs
must

$-$ respect the \TeX\ Live  conventions for command line parameters,\par
$-$ find its input files using the \.{kpathsearch} library, and\par
$-$ implement \TeX\ primitives to support 
\def\LaTeX{L\kern-.36em\raise.3ex\hbox{\sc A}\kern-.15em\TeX}%
\LaTeX.

\noindent Naturally, the functions that follow here are taken, with small
modifications, from the \TeX\ Live sources. What is
added here, or rather subtracted here, are the parts that are specific
to some of the \TeX\ engines included in \TeX\ Live. New is also
that the code is presented in literate programming style.

The code that follows is organized in three parts.
Some code for k\TeX\ must come before 
the definition of \TeX's macros because
it uses include files containing identifiers that are in conflict
with \TeX's macros. The remaining two parts are first auxiliar functions and then 
those functions that are called from the ``classic'' \TeX\ code.

@p @<k\TeX\ auxiliar functions@>@;
   @<k\TeX\ functions@>@;

@ Let's begin with the beginning: the command line.
To see how a command line is structured, we first look at the
help text that is displayed if the user asks for it (or if \TeX\
decides that the user needs it). The help text is produced by the
function |usage_help|.
\def\SP{ }% disable visible spaces in strings.

@<k\TeX\ auxiliar functions@>=
static void usage_help(void)
{@+@<explain the command line@>@;
  @<explain the options@>@;
  fprintf(stderr,"\nEmail bug reports to ruckert@@cs.hm.edu.\n");
  exit(0);
}

@ The command line commes in three slightly different versions:

@<explain the command line@>=
  fprintf(stderr,@/
    "Usage: %s [OPTION]... [TEXNAME[.tex]] [COMMANDS]\n"@/
    "   or: %s [OPTION]... \\FIRST-LINE\n"@/
    "   or: %s [OPTION]... &FMT ARGS\n\n",@/
    argv[0],argv[0],argv[0]);@/
  fprintf(stderr,@/
    "  Run TeX on TEXNAME, creating TEXNAME.dvi.\n"@/
    "  Any remaining COMMANDS are processed\n"@/
    "  as TeX input after TEXNAME is read.\n"@/
    "  If the first line of TEXNAME starts with %%&FMT, and FMT is\n"@/
    "  an existing .fmt file, use it. Else use `NAME.fmt', where\n"@/
    "  NAME is the program invocation name.\n"@/
    "\n"@/
    "  Alternatively, if the first non-option argument begins\n"@/
    "  with a backslash, interpret all non-option arguments as\n"@/
    "  a line of TeX input.\n"@/
    "\n"@/
    "  Alternatively, if the first non-option argument begins\n"@/
    "  with a &, the next word is taken as the FMT to read,\n"@/
    "  overriding all else.  Any remaining arguments are\n"@/
    "  processed as above.\n"@/
    "\n"@/
    "  If no arguments or options are specified, prompt for input.\n"@/
    "\n");

@ Here is the list of possible options and their explanation:

@<explain the options@>=
  fprintf(stderr,
  "Options:\n"@/
  " -help                 "@/
  @t\qquad@>"\t display this help and exit\n"@/
  " -version              "@/
  @t\qquad@>"\t output version information and exit\n"@/
  " -etex                 "@/
  @t\qquad@>"\t enable e-TeX extensions\n"@/
  " -ini                  "@/
  @t\qquad@>"\t be initex for dumping formats; this is\n"@/
  @t\qquad@>"\t\t\t also true if the program name is `kinitex'\n"@/
  " -progname=STRING      "@/
  @t\qquad@>"\t set program (and fmt) name to STRING\n"@/
  " -fmt=FMTNAME          "@/
  @t\qquad@>"\t use FMTNAME instead of program name or a %%& line\n"@/
  " -output-directory=DIR "@/
  @t\qquad@>"\t use existing DIR as the directory to write files to\n"@/
  " -jobname=STRING       "@/
  @t\qquad@>"\t set the TeX \\jobname to STRING\n"@/
   " [-no]-mktex=FMT       "@/
  @t\qquad@>"\t disable/enable mktexFMT generation (FMT=tex/tfm/fmt)\n"@/
  " -interaction=STRING   "@/
  @t\qquad@>"\t set interaction mode (STRING=batchmode/\n"@/
  @t\qquad@>"\t\t\t nonstopmode/scrollmode/errorstopmode)\n"@/
  " -kpathsea-debug=NUMBER"@/
  @t\qquad@>"\t set path searching debugging flags according\n"@/
  @t\qquad@>"\t\t\t to the bits of NUMBER\n"@/
  " [-no]-parse-first-line"@/
  @t\qquad@>"\t disable/enable parsing of the first line of\n"@/
  @t\qquad@>"\t\t\t the input file\n"@/
  " [-no]-file-line-error"@/
  @t\qquad@>"\t Disable/Enable file:line:error style\n"@/
  " -cnf-line=STRING"@/
  @t\qquad@>"\t Process STRING like a line in texmf.cnf\n"@/
);


@ The processing of command line options is controlled by the |long_options| 
\def\SP{{\tt\char`\ }}% restore the visible space in a string
array. Each entry in this array contains first the name of the option,
then a flag that tells whether the option takes an argument or not, 
then the (optional) address of a flag variable that gets set
by this option, and finally the value to store in the flag variable.

Besides the flag variables that occur in the table,
a few string variables may be set using the options.
The following is a complete list of these variables---except
for the the |interaction| variable of \TeX.
Flag variables are initialized with |-1| to indicate an undefined value;
string variables are initialized with |NULL|.

@<Global...@>=
static int iniversion=0;
static int etexp=0;
static int parsefirstlinep=-1;
static int filelineerrorstylep=-1;
static const char *user_progname=NULL, *output_directory=NULL, *c_job_name=NULL;
static char *dump_name=NULL;@#

static struct option long_options[] = {@/
      { "help",                      0, 0, 0 },@/
      { "version",                   0, 0, 0 },@/
      { "interaction",               1, 0, 0 },@/
      { "mktex",                     1, 0, 0 },@/
      { "no-mktex",                  1, 0, 0 },@/
      { "kpathsea-debug",            1, 0, 0 },@/
      { "progname",                  1, 0, 0 },@/
      { "fmt",                       1, 0, 0 },@/
      { "output-directory",          1, 0, 0 },@/
      { "jobname",                   1, 0, 0 },@/
      { "cnf-line",                  1, 0, 0 },@/
      { "ini",                       0, &iniversion, 1 },@/
      { "etex",                      0, &etexp, 1 },@/
      { "parse-first-line",          0, &parsefirstlinep, 1 },@/
      { "no-parse-first-line",       0, &parsefirstlinep, 0 },@/
      { "file-line-error",           0, &filelineerrorstylep, 1 },@/
      { "no-file-line-error",        0, &filelineerrorstylep, 0 },@/
      { 0, 0, 0, 0 }@+}@+;


@ Parsing the command line options is accomplished with the
|parse_options| function which in turn uses the |getopt_long_only|
function from the \CEE/ library. This function returns 0 and sets the
|option_index| parameter to the option found, or it returns $-1$ if
the end of all options is reached.
@<k\TeX\ functions@>=
static void parse_options (int argc, char *argv[])
{@+ while (true) {
    int option_index;
    int g = getopt_long_only (argc, argv, "+", long_options, &option_index);
    if (g==0)
    { @<handle the option at |option_index|@>@;@+ }
    else if (g == -1) return;
  }
}

@ To handle the options, we compare the name at the given |option_index| with
the different option names. This is not a very efficient method, but the
impact is low and it's simple to write.

Comparing the name of the argument with the |name| field in the |option| 
structure is done in the auxiliar function |argument_is|.
Unfortunately the |name| field is in conflict with the |name| macro
defined by \TeX. To avoid the conflict, the |argument_is| function 
goes just after the \.{kpathsea.h} header file that defines
the option structure.


@<Header files and function declarations@>=
#include <kpathsea/kpathsea.h>
static int argument_is(struct option *opt, char * s)
{@+ return STREQ(opt->name, s); @+}
#define ARGUMENT_IS(S) argument_is(long_options+option_index,S)

@ Now we can handle the first two options:
@<handle the option at |option_index|@>=
if (ARGUMENT_IS("help")) usage_help();        
else if (ARGUMENT_IS("version")){@+
       printf("Version 0.0\n");
       exit(0);@+
} 


@ The ``interaction'' option sets \TeX's |interaction| variable
based on its string argument contained in the |optarg| variable.

@<handle the option at |option_index|@>=
else @+if (ARGUMENT_IS ("interaction"))@t\2@> {
      if (STREQ (optarg, "batchmode"))        interaction = batch_mode;
      else if (STREQ (optarg, "nonstopmode")) interaction = nonstop_mode;
      else if (STREQ (optarg, "scrollmode"))  interaction = scroll_mode;
      else if (STREQ (optarg, "errorstopmode")) interaction = error_stop_mode;
      else WARNING1 ("Ignoring unknown argument `%s' to --interaction", optarg);
    } 

@ The next two options pass the string argument to the
\.{kpathsearch} library.

@<handle the option at |option_index|@>=
else @+if (ARGUMENT_IS ("mktex")) @t\2@> kpse_maketex_option (optarg, true);
else if (ARGUMENT_IS ("no-mktex")) kpse_maketex_option (optarg, false);


@ To debug the searching done by the \.{kpathsearch} library, 
the following otion can be used. 
The argument value 3 is a good choice to start with.

@<handle the option at |option_index|@>=
else @+if (ARGUMENT_IS ("kpathsea-debug")) @t\2@>
      kpathsea_debug |= atoi (optarg);


@ The remaining options either set a flag and are handled by 
the |getopt_long_only| function or they take a string argument 
and assign it to the corresponding string variable.

@<handle the option at |option_index|@>=
else @+if (ARGUMENT_IS ("progname"))@t\2@>  
  user_progname = normalize_quotes(optarg,"program name");
else if (ARGUMENT_IS ("fmt"))  
  dump_name = normalize_quotes(optarg,"format name");
else if (ARGUMENT_IS ("output-directory"))
  output_directory = normalize_quotes(optarg,"output directory");
else if (ARGUMENT_IS ("jobname"))
  c_job_name = normalize_quotes (optarg, "job name");
else if (ARGUMENT_IS ("cnf-line"))
  add_cnf_line(optarg);

@ When string arguments specify files or directories, 
special care is needed if arguments are quoted and/or contain spaces.
The function |normalize_quotes| makes shure that arguments containing 
spaces get quotes around them and it checks for unballanced quotes.


@<k\TeX\ auxiliar functions@>=
static char *normalize_quotes (const char *nom, const char *mesg)
{@+ int quoted = false;
    int must_quote = (strchr(nom, ' ') != NULL);
    char *ret = xmalloc(strlen(nom)+3);/*room for two quotes and NUL*/
    char *p=ret;
    const char *q;

    if (must_quote)
        *p++ = '"';
    for (q = nom; *q; q++)
        if (*q == '"')
            quoted = !quoted;@+
        else
            *p++ = *q;

    if (must_quote)
        *p++ = '"';
    *p = '\0';
    if (quoted) {
        fprintf(stderr, "! Unbalanced quotes in %s %s\n", mesg, nom);
        exit(1);
    }
    return ret;
}

@ With the \.{-cnf-line} option it is possible to specify 
a line of text as if this line were part of \TeX's configuration
file---even taking precedence over conflicting lines in the configuration
file. For example it is possible to change \TeX's
\.{TEXINPUTS} variable by saying \.{--cnf-line=TEXINPUTS=/foo}.
The configuration lines are temporarily stored in the variable
|cnf_lines| and counted in |cnf_count| because we can send them
to the \.{kpathsearch} library only after the library has been
initialized sufficiently.

@<Global...@>=
char **cnf_lines=NULL;
int cnf_count=0;

@ The function  |add_cnf_line| stores the 
given command line argument in the variable |cnf_lines|.

@<k\TeX\ auxiliar functions@>=
static void  add_cnf_line(char *arg)
{@+ cnf_count++;
  cnf_lines=xrealloc(cnf_lines,sizeof(char*)*cnf_count);
  cnf_lines[cnf_count-1]=arg;
}
 
@ To activate the configuration lines they are passed to the
 \.{kpathsearch} library.

@<activate configuration lines@>=
#if 0 /* this function does not exists allways */
{@+ int i;
  for (i=0;i<cnf_count;i++)
    kpathsea_cnf_line_env_progname (kpse_def, cnf_lines[i]);
  free(cnf_lines);
}
#endif

@ After we are done with the options, 
we inform the \.{kpathsearch} library about the program name.
This is an important piece of information for the library
because the library serves quite different programs
and its behaviour can be customized for each program 
using configuration files. 
After the program and engine name is set,
the library is ready to use.
@<set the program and engine name@>=
if (!user_progname)
    user_progname = dump_name;
#if defined(WIN32)
  if (user_progname)
    kpse_reset_program_name (user_progname);
#else
  kpse_set_program_name (argv[0], user_progname);
#endif
  xputenv ("engine", "hitex");@/


@ After the options, the command line usually continues with the name of the input
file. Getting a hold of the input file name can be quite complicated,
and we will have the first opportunity to use the \.{kpathsearch} library.

We start by looking at the first argument after the options: 
If it does not start with an
``\.{\AM}'' and neither with a ``\.{\BS}'', it's a simple file name.
Under Windows, however, filenames might start with a drive letter
followed by a colon and a ``\.{\BS}'' which is used to separate
directory names.  Finally, if the filename is a quoted string, we need
to remove the quotes before we use the \.{kpathsearch} library to find it
and reattach the quotes afterwards.

@<k\TeX\ auxiliar functions@>=
#ifdef WIN32
static void clean_windows_filename(char *filename)
{@+if (strlen (filename) > 2 && isalpha (filename[0]) &&
     filename[1] == ':' && filename[2] == '\\')
   { char *pp;
     for (pp = filename; *pp; pp++) 
       if (*pp == '\\')
         *pp = '/';
   }
}
#endif

static char *find_file(char *fname, kpse_file_format_type t, int mx)
{@+char *filename;
  int  final_quote=(int)strlen(fname)-1;
  int quoted = final_quote>1 && fname[0] == '"' && fname[final_quote]=='"';
  if (quoted) {
        /* Overwrite last quote and skip first quote. */
        fname[final_quote] = '\0';
        fname++;
  }
  filename = kpse_find_file(fname, t, mx);
  if (quoted) {
        /* Undo modifications */
        fname--;
        fname[final_quote] = '"';
  }
  return filename;
}

static char *get_input_file_name (void)
{@+
  char *input_file_name = NULL;

  if (argv[optind] && argv[optind][0] != '&' && argv[optind][0] != '\\') 
 { 
#ifdef WIN32
    clean_windows_filename(argv[optind]);
#endif
    argv[optind] = normalize_quotes(argv[optind], "input file");
    input_file_name=find_file(argv[optind],kpse_tex_format, false);
  }
  return input_file_name;
}


@ After we called |get_input_file_name|, we might need to look at |argv[argc-1]|
in case we run under Windows.

@<set the input file name@>=
  main_input_file = get_input_file_name ();

#ifdef WIN32
  /* Were we given a simple filename? */
  if (main_input_file == NULL) {
    char *file_name=argv[argc-1];
    if (file_name && file_name[0] != '-' && file_name[0] != '&' && file_name[0] != '\\') {
    clean_windows_filename(file_name);
    file_name = normalize_quotes(file_name, "argument");
    main_input_file=find_file(file_name,kpse_tex_format, false);
    argv[argc-1] = file_name;
    }
  }
#endif


@ After we have an input file, we make an attempt at filling in options
from the {\tt texmf.cfg} file.

@<set defaults from the {\tt texmf.cfg} file@>=
  if (filelineerrorstylep < 0) 
    filelineerrorstylep = texmf_yesno ("file_line_error_style");
  if (parsefirstlinep < 0)
    parsefirstlinep = texmf_yesno ("parse_first_line");@/@t~@>

@ We needed:
@<k\TeX\ auxiliar functions@>=
static int texmf_yesno(const char *var)
{@+
  char *value = kpse_var_value (var);
  return value && (*value == 't' || *value == 'y' || *value == '1');
}

@ We need a stack, matching the |line_stack| that 
contains the source file names;
we postpone k\TeX\'s |full_source_filename_stack| to a later time.

@<Global...@>=
int @!source_filename_stack0[max_in_open], *const @!source_filename_stack = @!source_filename_stack0-1;

@ The function |print_file_line|
prints ``file:line:error'' style messages.  It looks for a
filename in |source_filename_stack|, and if it fails to find
one falls back on the ``non-file:line:error'' style.

@<Basic printing...@>=
static void print_file_line(void)
{@+int level=in_open;
  while (level>0 && source_filename_stack[level]==0) level--;
  if (level==0) print_nl("! ");
  else
  { print_nl(""); printn(source_filename_stack[level]); print_char(':');
    if (level==in_open) print_int(line);
    else print_int(line_stack[level]);
    print(": ");
  }
} 

@ Most of the time \TeX\ is not running as \.{initex} or \.{virtex},
but it runs with a format file preloaded.
To set the format name, we first check if the format name was given on the
command line with an ``\.{\AM}'' prefix,
second we might check the first line of the input file,
and last, we check if the program is an initex or virtex program.
  
If we still don't have a format,
we use a plain format if running as a virtex, otherwise the program
name is our best guess.
There is no need to check for an extension, because the 
\.{kpathsearch} library will take care of that.
We store the format file name in  |dump_name| which is 
used in the function |w_open_in| below.

@<set the format name@>=
if (parsefirstlinep && !dump_name)
  parse_first_line (main_input_file);
if (!main_input_file && argv[1] && argv[1][0] == '&') 
  dump_name = argv[1] + 1;
if (strcmp(kpse_program_name, "kinitex")==0) iniversion = true;
else if (strcmp(kpse_program_name, "kvirtex")==0 && !dump_name)
  dump_name = "ktex";
if (!dump_name) 
  dump_name = kpse_program_name;
if (!dump_name) 
{ fprintf(stderr,"Unable to determine format name\n");
  exit(1);
}
if (etexp && !iniversion) 
{ fprintf(stderr,"-etex requires -ini\n");
  exit(1);
}

@ Here is the function |parse_first_line |. It searches the first line of
the file for a \TeX\ comment of the form ``\%\AM format''\footnote{${}^1$}{The idea of using this format came from Wlodzimierz Bzyl.}.
If found, we will use the format given there.

@<k\TeX\ functions@>=
static void parse_first_line (char*filename)
{@+ FILE *f=NULL;
  if (filename==NULL) return;
  f = open_in(filename,kpse_tex_format,"r");
  if (f!=NULL)
  { char *r,*s,*t=read_line (f);
    xfclose (f, filename);
    if (t==NULL) return;
    s=t;
    if (s[0] == '%' && s[1] == '&') {
      s = s+2;
      while ( ISBLANK(*s)) ++s;
      r=s;
      while (*s!=0 && *s!=' ' && *s!='\r' && *s!='\n') s++;
      *s=0;
      if (dump_name==NULL) {
        char *f_name = concat (r, ".fmt");
        char *d_name = kpse_find_file (f_name,kpse_fmt_format, false);
        if (d_name && kpse_readable_file (d_name)) {
            dump_name = xstrdup (r);
            kpse_reset_program_name (dump_name);
          }
          free (f_name);
        }
    }
    free (t);
  }
}


@ The \TeX\ Live infrastructure is able to generate format files, 
font metric files, 
and even some tex files, if required.

@<enable the generation of input files@>=
  kpse_set_program_enabled (kpse_tfm_format, MAKE_TEX_TFM_BY_DEFAULT,
                            kpse_src_compile);@/
  kpse_set_program_enabled (kpse_tex_format, MAKE_TEX_TEX_BY_DEFAULT,
                            kpse_src_compile);@/
  kpse_set_program_enabled (kpse_fmt_format, MAKE_TEX_FMT_BY_DEFAULT,
                            kpse_src_compile);@/


@ In the old days, \TeX\ was a \PASCAL\ program, and standard \PASCAL\
did say nothing about a command line. So \TeX\ would open the terminal
file for input and read all the information from the terminal. 
If you don't give \TeX\ command line arguments, this is still true today.
In our present time, people got so much used to control the behaviour
of a program using command line arguments---especially when writing
scripts---that k\TeX\ allows the specification of commands on
the command line which \TeX\ would normally expect on
the first line of its terminal input. 

So our next task is
writing a function to add the remainder of the 
command line to \TeX's input buffer.
The main job is done by the |input_add_str| function 
which duplicates part of the |input_ln| function.
Further it skips initial spaces and replaces trailing spaces
and line endings by a single space.

@<k\TeX\ auxiliar functions@>=
static void input_add_char(unsigned int c)
{@+ if (last >= max_buf_stack) 
  {@+max_buf_stack=last+1;
     if (max_buf_stack==buf_size) 
       @<Report overflow of the input buffer, and abort@>;
  } 
  buffer[last]=xord[c];incr(last);
}    

static void input_add_str(const char *str)
{@+ int prev_last;
  while (*str==' ') str++;
  prev_last=last;
  while (*str!=0) 

input_add_char(*str++);
  for (--last; last >= first ; --last) 
  { char c =buffer[last];
    if ((c) != ' ' && (c) != '\r' && (c) != '\n') break;
  }
  last++;
  if (last>prev_last) input_add_char(' ');
}

static int input_command_line(void)
{@+ 
last=first;
while (optind < argc) input_add_str(argv[optind++]);
loc=first;
return (loc < last);
@+ }



@ When we open an output file, there is usually no searching necessary.
In the best case, we have an absolute path and can open it.
If the path is relative, we try in this order:
the |file_name| prefixed by the |output_directory|,
the |file_name| as is, and
the |file_name| prefixed with the environment variable |TEXMFOUTPUT|.

If we were successful with one of the modified names, we update
|name_of_file|.

@<k\TeX\ functions@>=
static FILE *open_out(const char *file_name, const char *file_mode)
{@+ FILE *f=NULL;
  char *new_name=NULL;
  int absolute = kpse_absolute_p(file_name, false);
  if (absolute)
    return fopen(file_name,file_mode);
  if (output_directory) 
  { new_name = concat3(output_directory, DIR_SEP_STRING, file_name);
    f = fopen(new_name,file_mode);
    if (f==NULL) {@+ free(new_name); new_name=NULL;@+}
  }
  if (f==NULL)  
    f = fopen(file_name,file_mode);
  if (f==NULL)
  { const char *texmfoutput = kpse_var_value("TEXMFOUTPUT");
    if (texmfoutput!=NULL && texmfoutput[0]!=0)
    { new_name = concat3(texmfoutput, DIR_SEP_STRING, file_name);@/
      f = fopen(new_name,file_mode);
      if (f==NULL) {@+ free(new_name); new_name=NULL; @+}
    }  
  }
  if (f!=NULL && new_name!=NULL) 
    update_name_of_file(new_name,(int)strlen(new_name));
  if (new_name!=NULL) free(new_name);
  return f;
}

static bool a_open_out(alpha_file *f)
   /*open a text file for output*/
{@+f->f=open_out((char*)name_of_file+1,"w");
  return f->f!=NULL && ferror(f->f)==0;@+
}

static bool b_open_out(byte_file *f)
   /*open a binary file for output*/
{@+f->f=open_out((char *)name_of_file+1,"wb");
return f->f!=NULL && ferror(f->f)==0;@+
}

#ifdef @!INIT
static bool w_open_out(word_file *f)
   /*open a word file for output*/
{@+f->f=open_out((char *)name_of_file+1,"wb");
   return f->f!=NULL && ferror(f->f)==0;@+
}
#endif

@ Format file names must be scanned
before \TeX's string mechanism has been initialized. 
The function |update_name_of_file| will set |name_of_file| from
a \CEE/ string. 

We dare not give error messages here, since \TeX\ calls this routine before
the |error| routine is ready to roll. Instead, we simply drop excess characters,
since the error will be detected in another way when a strange file name
isn't found.
@^system dependencies@>

@<k\TeX\ auxiliar functions@>=
static void update_name_of_file(char *@!s, int @!k)
{@+ int j;
  if (k <= file_name_size) name_length=k;@+
  else name_length=file_name_size;
  for (j=0; j<name_length;j++)  name_of_file[j+1]=xchr[(int)s[j]];
  name_of_file[name_length+1]=0;
}

@ In standard \TeX, the |reset| macro is used to open input files.
The \.{kpathsearch} library uses different search pathes for
different types of files and therefore different functions are needed to
open these files. The common code is in the function |open_in|.


@<k\TeX\ auxiliar functions@>=
static FILE*open_in(char*filename,kpse_file_format_type t,const char*rwb)
{@+ char *fname=NULL;
  FILE*f= NULL;
  fname= find_file(filename,t,true);
  if(fname!=NULL)
  {@+
    f= fopen(fname,rwb);
    free(fname);@+
  }
  return f;
}

@ @<k\TeX\ functions@>=
static bool a_open_in(alpha_file *f) /*open a text file for input*/ 
{@+f->f= open_in((char *)name_of_file+1,kpse_tex_format,"r");
   if (f->f!=NULL) get(*f);
   return f->f!=NULL && ferror(f->f)==0;
}

static bool b_open_in(byte_file *f)   /*open a binary file for input*/ 
{@+f->f= kpse_open_file((char *)name_of_file+1,kpse_tfm_format);
   if (f->f!=NULL) get(*f);
   return f->f!=NULL && ferror(f->f)==0;
} 


@ \TeX's |open_fmt_file| function will call the following function
either with the name of a format file as given with an ``\.{\AM}'' 
prefix in the input or with |NULL| if no such name was specified.
The function will try |dump_name| as a last resort before returning |NULL|.



@<k\TeX\ functions@>=
static bool w_open_in(word_file *f)   /*open a word file for input*/ 
{@+
  f->f=NULL;
  if (name_of_file[1]!=0)
    f->f= open_in((char*)name_of_file+1,kpse_fmt_format,"rb");
  if (f->f!=NULL) get(*f);
  return f->f!=NULL && ferror(f->f)==0;
} 

static bool open_fmt_file(void)
{@+int j=loc;
  if (buffer[loc]=='&')
  {@+incr(loc);j=loc;buffer[last]=' ';
    while (buffer[j]!=' ') incr(j);
    update_name_of_file((char *)buffer+loc,j-loc);
    if (w_open_in(&fmt_file)) goto found;
  }
  update_name_of_file(dump_name,(int)strlen(dump_name));
  if (w_open_in(&fmt_file)) goto found;
  name_of_file[1]=0;
  wake_up_terminal;
  wterm_ln("I can't find a format file!");
  return false;
found: loc=j;return true;
}



@ All the code that we have seen so far commes together in the function |main_init|
which is the first function called in the main program of a \TeX\ engine 
belonging to \TeX\ Live. Before doing so, we make copies of
argument count and argument vector putting them in global variables.
@<Global...@>=
static char **argv;
static int argc;

@ @<k\TeX\ functions@>=
static void main_init(int ac, char *av[])
{@+
  char* main_input_file;
  argc = ac;
  argv = av;
  interaction = error_stop_mode;@/
  @<parse options@>@;
  @<set the program and engine name@>@;
  @<activate configuration lines@>@;
  @<set the input file name@>@;
  @<set defaults from the {\tt texmf.cfg} file@>@;
  @<set the format name@>@;
  @<enable the generation of input files@>@;
}


@ Before we can call the |parse_options| function,
we might need some special preparations for Windows.

@<parse options@>=
#if defined(WIN32)
{@+ char* enc;
  kpse_set_program_name (argv[0], NULL);
  enc = kpse_var_value("command_line_encoding");
  get_command_line_args_utf8(enc, &argc, &argv);@/
  parse_options (argc, argv);
}
#else
  parse_options (ac, av);
#endif

@ Adding prototypes for the functions just defined allows \TeX\
to call them.

@<Header files and function declarations@>=

static int input_command_line(void);
static void main_init(int ac, char *av[]);

@ We conclude this chapter using \.{time.h} to provide a function 
that is used to initialize 
\TeX's date and time information. Because |time| is one of \TeX's
macros, we add the function |tl_now| that wraps the call to |time|
(and |gmtime|) for later use in |fix_date_and_time|.

@<Header files and function declarations@>=
#include <time.h>
static struct tm *tl_now(void)
{@+ time_t t=time(NULL);
  return gmtime(&t);
}


@* Index.
@z